//******************************************************************************
//
// Simple MIDI Library / SMTrack
//
// gbNNX
//
// Copyright (C) 2010 WADA Masashi. All Rights Reserved.
//
//******************************************************************************

#include "StdAfx.h"
#include "YNBaseLib.h"
#include "SMTrack.h"
#include "SMEventMIDI.h"

using namespace YNBaseLib;

namespace SMIDILib {


//******************************************************************************
// RXgN^
//******************************************************************************
SMTrack::SMTrack(void)
 : m_List(sizeof(SMDataSet), 1000)
{
}

//******************************************************************************
// fXgN^
//******************************************************************************
SMTrack::~SMTrack(void)
{
	Clear();
}

//******************************************************************************
// f[^NA
//******************************************************************************
void SMTrack::Clear()
{
	SMExDataMap::iterator exdataitr;

	m_List.Clear();
	
	for (exdataitr = m_ExDataMap.begin(); exdataitr != m_ExDataMap.end(); exdataitr++) {
		delete [] (exdataitr->second);
	}
	m_ExDataMap.clear();
	
	return;
}

//******************************************************************************
// f[^Zbgǉ
//******************************************************************************
int SMTrack::AddDataSet(
		unsigned long deltaTime,
		SMEvent* pEvent,
		unsigned char portNo
	)
{
	int result = 0;
	unsigned long index = 0;
	unsigned long blockNo = 0;
	unsigned long blockIndex = 0;
	unsigned char* pBlock = NULL;
	unsigned char* pExData = NULL;
	SMDataSet dataSet;

	index = m_List.GetSize();

	//f[^Zbg쐬
	ZeroMemory(&dataSet, sizeof(SMDataSet));
	dataSet.deltaTime = deltaTime;
	dataSet.eventData.type   = pEvent->GetType();
	dataSet.eventData.status = pEvent->GetStatus();
	dataSet.eventData.meta   = pEvent->GetMetaType();
	dataSet.eventData.size   = pEvent->GetDataSize();
	dataSet.portNo = portNo;

	//Cxgf[^4byteȓȂ\̓Ɋi[
	if (pEvent->GetDataSize() <= 4) {
		memcpy(&(dataSet.eventData.data), pEvent->GetDataPtr(), pEvent->GetDataSize());
	}
	//ȊO͕ʓrq[vɕێă}bvŊǗ
	else {
		try {
			pExData = new unsigned char[pEvent->GetDataSize()];
		}
		catch (std::bad_alloc) {
			result = YN_SET_ERR("Could not allocate memory.", pEvent->GetDataSize(), 0);
			goto EXIT;
		}
		memcpy(pExData, pEvent->GetDataPtr(), pEvent->GetDataSize());
		m_ExDataMap.insert(SMExDataMapPair(index, pExData));
		pExData = NULL;
	}

	result = m_List.AddItem(&dataSet);
	if (result != 0) goto EXIT;

EXIT:;
	delete [] pExData;
	return result;
}

//******************************************************************************
// f[^Zbg擾
//******************************************************************************
int SMTrack::GetDataSet(
		unsigned long index,
		unsigned long* pDeltaTime,
		SMEvent* pEvent,
		unsigned char* pProtNo
	)
{
	int result = 0;
	unsigned long blockNo = 0;
	unsigned long blockIndex = 0;
	unsigned char* pBlock = NULL;
	unsigned char* pEventData = NULL;
	SMDataSet dataSet;
	SMExDataMap::iterator exdataitr;

	result = m_List.GetItem(index, &dataSet);
	if (result != 0) goto EXIT;

	//f^^C
	if (pDeltaTime != NULL) {
		*pDeltaTime = dataSet.deltaTime;
	}

	//Cxgf[^ʒu
	if (dataSet.eventData.size <= 4) {
		pEventData = dataSet.eventData.data;
	}
	else {
		exdataitr = m_ExDataMap.find(index);
		if (exdataitr == m_ExDataMap.end()) {
			result = YN_SET_ERR("Program error.", index, 0);
			goto EXIT;
		}
		pEventData = exdataitr->second;
	}

	//Cxgf[^o^
	if (pEvent != NULL) {
		result = pEvent->SetData(
						dataSet.eventData.type,
						dataSet.eventData.status,
						dataSet.eventData.meta,
						pEventData,
						dataSet.eventData.size
					);
		if (result != 0) goto EXIT;
	}

	//|[gԍ
	if (pProtNo != NULL) {
		*pProtNo = dataSet.portNo;
	}

EXIT:;
	return result;
}

//******************************************************************************
// Rs[
//******************************************************************************
unsigned long SMTrack::GetSize()
{
	return m_List.GetSize();
}

//******************************************************************************
// Rs[
//******************************************************************************
int SMTrack::CopyFrom(
		SMTrack* pSrcTrack
	)
{
	int result = 0;
	unsigned long index = 0;
	unsigned long deltaTime = 0;
	SMEvent event;
	unsigned char portNo = 0;

	//TODO: CeWFgȃRs[ɂ
	
	if (pSrcTrack == NULL) {
		result = YN_SET_ERR("Program error.", 0, 0);
		goto EXIT;
	}
	
	//gRs[Ȃ牽Ȃ
	if (pSrcTrack == this) {
		goto EXIT;
	}
	
	Clear();
	
	for (index = 0; index < pSrcTrack->GetSize(); index++) {
		result = pSrcTrack->GetDataSet(index, &deltaTime, &event, &portNo);
		if (result != 0) goto EXIT;
		
		result = AddDataSet(deltaTime, &event, portNo);
		if (result != 0) goto EXIT;
	}
	
EXIT:;
	return result;
}

//******************************************************************************
// m[gXg擾
//******************************************************************************
int SMTrack::GetNoteList(
		SMNoteList* pNoteList
	)
{
	int result = 0;
	unsigned long index = 0;
	unsigned long deltaTime = 0;
	unsigned long totalTime = 0;
	unsigned char portNo = 0;
	unsigned long key = 0;
	SMNoteMap noteMap;
	SMNoteMap::iterator itr;

	if (pNoteList == NULL) {
		result = YN_SET_ERR("Program error.", 0, 0);
		goto EXIT;
	}

	pNoteList->Clear();

	for (index = 0; index < GetSize(); index++) {
		
		SMEvent event;
		result = GetDataSet(index, &deltaTime, &event, &portNo);
		if (result != 0) goto EXIT;

		totalTime += deltaTime;

		if (event.GetType() == SMEvent::EventMIDI) {

			SMEventMIDI midiEvent;
			midiEvent.Attach(&event);

			//m[gI
			if (midiEvent.GetChMsg() == SMEventMIDI::NoteOn) {
				//o^ς݂łȂ΃}bvɓo^
				key = _GetNoteKey(portNo, midiEvent.GetChNo(), midiEvent.GetNoteNo());
				itr = noteMap.find(key);
				if (itr == noteMap.end()) {
					SMNote note;
					ZeroMemory(&note, sizeof(SMNote));
					note.portNo = portNo;
					note.chNo = midiEvent.GetChNo();
					note.noteNo = midiEvent.GetNoteNo();
					note.velocity = midiEvent.GetVelocity();
					note.startTime = totalTime;
					noteMap.insert(SMNoteMapPair(key, note));
				}
			}
			//m[gIt
			if (midiEvent.GetChMsg() == SMEventMIDI::NoteOff) {
				//}bvo^ς݂̂͂Ȃ̂Ń}bv폜ăm[gXgɒǉ
				key = _GetNoteKey(portNo, midiEvent.GetChNo(), midiEvent.GetNoteNo());
				itr = noteMap.find(key);
				if (itr != noteMap.end()) {
					SMNote note;
					note = itr->second;
					note.endTime = totalTime;
					pNoteList->AddNote(note);
					noteMap.erase(itr);
				}
			}
		}
	}

	//m[gÎ܂܏Iꍇ͋؂m[gXgɒǉ
	for (itr = noteMap.begin(); itr != noteMap.end(); itr++) {
		SMNote note;
		note = itr->second;
		note.endTime = totalTime;
		pNoteList->AddNote(note);
	}

EXIT:;
	return result;
}

//******************************************************************************
// m[gL[擾
//******************************************************************************
unsigned long SMTrack::_GetNoteKey(
		unsigned char portNo,
		unsigned char chNo,
		unsigned char noteNo
	)
{
	return ((portNo << 16) | (chNo << 8) | noteNo);
}

} // end of namespace

